# pylint: disable=no-member

import datetime
import random
import uuid

from qiskit import Aer, QuantumCircuit, assemble
from qiskit.providers import JobStatus


def fake_circuit():
    circuit = QuantumCircuit(2, 2)
    circuit.h(0)
    circuit.cx(0, 1)
    circuit.measure([0, 1], [0, 1])

    return circuit


def fake_qobj():
    backend = Aer.get_backend("qasm_simulator")
    return assemble(fake_circuit(), backend)


def fake_job_dict():
    job_id = str(uuid.uuid4())
    qobj = fake_qobj()

    return {
        "job_id": job_id,
        "date_submit": datetime.datetime.now().isoformat(),
        "qobj": qobj.to_dict(),
        "status": random.choice([
            JobStatus.INITIALIZING.name,
            JobStatus.RUNNING.name
        ])
    }



import uuid

import pytest

from qctic.api import QCticAPI
from qctic.backend import QCticQasmSimulator
from qctic.job import QCticJob
from qctic.provider import QCticProvider
from tests.utils import fake_qobj


@pytest.fixture(scope="function")
def provider(httpserver):
    api_host = httpserver.url_for("/")
    api = QCticAPI(host=api_host)
    api.auth_token(str(uuid.uuid4()))
    return QCticProvider(api=api)


@pytest.fixture(scope="function")
def job(provider):
    backend = provider.get_backend(QCticQasmSimulator.NAME)
    job_id = str(uuid.uuid4())
    qobj = fake_qobj()
    return QCticJob(backend, job_id, qobj)



import logging
import os

from qiskit.providers import BaseProvider
from qiskit.providers.providerutils import filter_backends

from qctic.api import QCticAPI
from qctic.backend import (QCticQasmSimulator, QCticStatevectorSimulator,
                           QCticUnitarySimulator)

_logger = logging.getLogger(__name__)


class QCticProvider(BaseProvider):
    """Provider for CTIC Erwin backends."""

    ENV_HOST = "QCTIC_HOST"
    ENV_USERNAME = "QCTIC_USERNAME"
    ENV_PASSWORD = "QCTIC_PASSWORD"
    ENV_TOKEN = "JUPYTERHUB_API_TOKEN"

    def __init__(self, *args, **kwargs):
        self._api = kwargs.pop("api")

        if not self._api:
            raise Exception("Must set the api kwarg")

        super().__init__(args, kwargs)

        self._backends = [
            QCticQasmSimulator(provider=self),
            QCticStatevectorSimulator(provider=self),
            QCticUnitarySimulator(provider=self)
        ]

    @classmethod
    def from_env(cls, *args, **kwargs):
        host = os.getenv(cls.ENV_HOST, None)
        username = os.getenv(cls.ENV_USERNAME, None)
        password = os.getenv(cls.ENV_PASSWORD, None)
        token = os.getenv(cls.ENV_TOKEN, None)

        err_msg = "Missing Erwin environment variables: {}".format({
            cls.ENV_HOST: host,
            cls.ENV_USERNAME: username,
            cls.ENV_PASSWORD: password,
            cls.ENV_TOKEN: token
        })

        if not host:
            _logger.warning(err_msg)
            raise Exception(err_msg)

        api = QCticAPI(host=host)

        if username and password:
            api.auth_basic(username, password)
        elif token:
            api.auth_token(token)
        else:
            _logger.warning(err_msg)
            raise Exception(err_msg)

        return cls(api=api, *args, **kwargs)

    @property
    def api(self):
        return self._api

    def get_backend(self, name=None, **kwargs):
        return super().get_backend(name=name, **kwargs)

    def backends(self, name=None, filters=None, **kwargs):
        backends = self._backends

        if name:
            backends = [
                backend for backend in backends
                if backend.name() == name
            ]

        return filter_backends(backends, filters=filters, **kwargs)

    def __str__(self):
        return self.__class__.__name__



import setuptools

from qctic.__version__ import __version__

setuptools.setup(
    name="qctic",
    version=__version__,
    keywords="quantum simulator qiskit",
    author="Andres Garcia Mangas",
    author_email="andres.garcia@fundacionctic.org",
    description="Qiskit provider to interact with the CTIC Erwin quantum simulator platform",
    url="https://bitbucket.org/fundacionctic/erwin-qiskit",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License"
    ],
    entry_points={
        "console_scripts": [
            "qcticapi=qctic.cli:run_local_server"
        ]
    },
    python_requires=">=3.6",
    install_requires=[
        "qiskit-aer>=0.4.0,<1.0.0",
        "marshmallow>=3.0,<4.0",
        "tornado>=6.0,<7.0"
    ],
    extras_require={
        "dev": [
            "autopep8>=1.5,<2.0",
            "pylint>=2.0,<3.0",
            "rope>=0.16.0,<1.0",
            "pytest>=5.0,<6.0",
            "bumpversion>=0.5.3,<1.0",
            "pytest-httpserver>=0.3.0,<1.0",
            "coloredlogs>=14.0,<15.0",
            "pytest-cov>=2.8.1,<3.0",
            "pytest-asyncio>=0.10.0,<1.0",
            "notebook>=6.0,<7.0",
            "coloredlogs>=14.0,<15.0"
        ]
    }
)



import asyncio
import logging
import random
import time
from functools import wraps

from qiskit.providers import BaseJob, JobError, JobStatus
from qiskit.result import Result

_logger = logging.getLogger(__name__)


def _fetch_job_sync(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        fetch = kwargs.pop("fetch", True)

        if not args[0]._remote_job or fetch:
            args[0]._fetch_remote_job()

        return func(*args, **kwargs)

    return wrapper


def _fetch_job_async(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        fetch = kwargs.pop("fetch", True)

        if not args[0]._remote_job or fetch:
            await args[0]._fetch_remote_job_async()

        return await func(*args, **kwargs)

    return wrapper


async def _dummy_submit_task():
    _logger.debug((
        "Waiting on dummy submit task: "
        "Was job.submit() in backend.run() synchronous?"
    ))


class QCticJob(BaseJob):
    """A Qiskit job that is executed remotely in the CTIC Erwin simulation platform."""

    def __init__(self, backend, job_id, qobj, run_params=None):
        """Constructor.

        Args:
            backend (QCticQasmSimulator): Backend that contains this job.
            job_id (str): ID of the job.
            qobj (Qobj): Qobj that contains the experiment.
            run_params (dict): Optional dict of arbitrary arguments passed
                to the ``run`` method in the remote simulation platform.
        """

        super().__init__(backend, job_id)
        self._qobj = qobj
        self._remote_job = None
        self._fetch_time = None
        self._run_params = run_params
        self._submit_task = None

    @property
    def api(self):
        """QCticAPI: The API instance of this job's backend."""

        return self.backend().api

    @property
    def qobj(self):
        """Return the Qobj submitted for this job.

        Returns:
            Qobj: The Qobj submitted for this job.
        """

        return self._qobj

    @property
    def run_params(self):
        """Arguments passed to the ``run`` method in the remote simulation platform.

        Returns:
            dict: The arguments dict.
        """

        return self._run_params

    @property
    def submit_task(self):
        return self._submit_task if self._submit_task else _dummy_submit_task()

    @submit_task.setter
    def submit_task(self, val):
        if self._submit_task:
            raise ValueError("The submit task has already been defined")

        self._submit_task = val

    def _log_fetch(self):
        _logger.debug("Fetching job: %s", self.job_id())

    def _fetch_remote_job(self):
        self._log_fetch()
        self._remote_job = self.api.get_job_sync(self.job_id())
        self._fetch_time = time.time()

    async def _fetch_remote_job_async(self):
        self._log_fetch()
        self._remote_job = await self.api.get_job(self.job_id())
        self._fetch_time = time.time()

    def _result(self):
        error = self._remote_job.get("error", None)

        if error is not None:
            raise JobError(error)

        result_dict = self._remote_job.get("result", None)

        return Result.from_dict(result_dict) if result_dict else None

    def _status(self):
        return self._remote_job.get("status", None)

    @_fetch_job_sync
    def status(self):
        """Return the status of the job, one of the names of the ``JobStatus`` enum.

        Args:
            fetch (bool): (Optional) Fetch the status from the remote API.

        Returns:
            JobStatus: The name of a member of the JobStatus enum.
        """

        return self._status()

    def submit(self):
        """Submit the job to the backend for execution."""

        self.api.post_job_sync(self)

    def cancel(self):
        """Attempt to cancel the job."""

        self.api.cancel_job_sync(self.job_id())

    @_fetch_job_sync
    def result(self, poll=False, base_sleep=2, max_sleep=128, sleep_growth=2):
        """Return the results of the job.

        Args:
            fetch (bool): (Optional) Fetch the result from the remote API.

        Returns:
            Result: The results.

        Raises:
            JobError: If there was an error running the experiment on the remote simulator.
        """

        sleep = base_sleep

        while True:
            res = self._result()

            if res or not poll:
                return res

            time.sleep(sleep)
            sleep = min(sleep * sleep_growth, max_sleep)
            self._fetch_remote_job()

    @_fetch_job_async
    async def status_async(self):
        return self._status()

    async def submit_async(self):
        await self.api.post_job(self)

    async def cancel_async(self):
        await self.api.cancel_job(self.job_id())

    @_fetch_job_async
    async def result_async(self, poll=False, base_sleep=2, max_sleep=128, sleep_growth=2):
        sleep = base_sleep

        while True:
            res = self._result()

            if res or not poll:
                return res

            await asyncio.sleep(sleep)
            sleep = min(sleep * sleep_growth, max_sleep)
            await self._fetch_remote_job_async()
